#!/usr/bin/python
#
# fit - tools for fitting WLCs to force spectroscopy unfolding data
# and extracting unfolding forces.
#
# Copyright (C) 2008-2009, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.
"""Provide tools for fitting WLCs to unfolding force curves and
extracting unfolding forces."""

import numpy
from scipy.optimize import leastsq
import os.path

import quickhull2d # my extended version, see http://www.physics.drexel.edu/~wking/code/python/quickhull2d.py

KB = 1.3806503e-23 # J/K

def readFile(fid, cols=2, sep='\t', dtype=numpy.double, split_cols=True):
    """This is a low level wrapper around numpy.fromfile.  It works on
    text and binary files, but you need to know the number of columns
    in the file ahead of time.
    """
    data = numpy.fromfile(file=fid, dtype=dtype, sep=sep)
    rows = len(data)/cols
    data = data.reshape((rows,cols))
    if split_cols == True:
        col_list = []
        for i in range(cols):
            col_list.append(data[:,i])
        return col_list
    else:
        return data

class WLC (object):
    """Wormlike Chain polymer tension model.
    The current implementation uses Bustamante's interoplation formula.
    
    Bustamante, Marko, Siggia, and Smith.  "Entropic Elasticity of lambda-Phage DNA."
    Science, New Series, Vol. 265, No. 5178. (Sep. 9, 1994), pp. 1599--1600.
    """
    def __init__(self, p=0.1, L=1.0, T=298.15, fixedP=True, fixedL=False, fixedT=True):
        self.p = float(p)
        self.L = float(L)
        self.T = float(T)
        self.fixedP = fixedP
        self.fixedL = fixedL
        self.fixedT = fixedT
        self.indexFree()
    def __str__(self):
        return "<WLC p:%g L:%g T:%g>" % (self.p, self.L, self.T)
    def __repr__(self):
        return str(self)
    def indexFree(self):
        self.nparam = 0
        self._paramFree = []
        self._paramNames = 'pLT'
        if self.fixedP == False:
            self.nparam += 1
            self._paramFree.append('p')
        if self.fixedL == False:
            self.nparam += 1
            self._paramFree.append('L')
        if self.fixedT == False:
            self.nparam += 1
            self._paramFree.append('T')
    def setFree(self, paramArray):
        if not hasattr(paramArray, '__len__'):
            paramArray = [float(paramArray)] # single value need not be in an array
        assert len(paramArray) == self.nparam
        for paramName,value in zip(self._paramFree, paramArray):
            self.__setattr__(paramName, abs(float(value)))
    def WLC(self, x):
        x = x/self.L # scale by L
        if hasattr(x,  '__len__'): # x is an array
            ansArray = KB*self.T/self.p * (  0.25*( 1/(1-x)**2 - 1 ) + x  )
            for xnum,i in zip(x, range(len(x))):
                if xnum <= 0:
                    ansArray[i] = 0
                elif xnum >= 1:
                    ansArray[i] = float('infinity')
            return ansArray
        else: # x is not an array
            if x <= 0:
                return 0
            elif x >= 1:
                return float('infinity')
            else:
                return KB*self.T/self.p * (  0.25*( 1/(1-x)**2 - 1 ) + x  )
    def residual(self, p, y, x):
        self.setFree(p)
        Y = self.WLC(x)
        return Y-y
    def fit(self, xdata, ydata, paramGuess=None, textVerbose=False, plotVerbose=False):
        assert len(xdata) > 0, "No xdata!  Can't fit that."
        assert len(xdata) == len(ydata), "X and Y data must be same length, but %d != %d" % (len(xdata), len(ydata))
        self.indexFree()
        if paramGuess != None:
            self.setFree(paramGuess)
        pGuess = []
        for paramName in self._paramFree:
            assert self._paramNames.count(paramName) > 0
            pGuess.append(getattr(self, paramName))
        p,cov,info,mesg,ier = leastsq(self.residual, pGuess, args=(ydata, xdata), full_output=True, maxfev=10000)
        if textVerbose:
            print "Fitted params:",p
            print "Covariance mx:",cov
            print "Info:", info
            print "mesg:", mesg
            if ier == 1:
                print "Solution converged"
            else:
                print "Solution did not converge"
        self.setFree(p)
        if plotVerbose:
            import pylab
            pylab.figure()
            pylab.plot(xdata, ydata, 'k.')
            pylab.hold(True)
            ymodel = []
            for x in xdata:
                ymodel.append(self.WLC(x))
            pylab.plot(xdata, ymodel, 'r-')
            pylab.show()
        #assert ier == 1, mesg

def _testWLC(p=0.4e-9, L=28, T=298.15,
             pgf=0.367879, Lgf=3.14159, Tgf=1.0,
             pfix=True, Lfix=False, Tfix=True, randomScaleFactor=0.3,
             npoints=100, maxLfrac=0.8,
             textVerbose=False, plotVerbose=False):
    """Xgf is the "Guess Factor" for parameter x.  For example, pGuess = p*pgf
    The Guess Factor is only applied to fitted (unfixed) parameters.
    """
    from random import Random
    assert Lgf > maxLfrac, "Evaluation would have some x > L"
    print "Testing WLC"
    r = Random()
    wlc = WLC(p=p, L=L, T=T)
    print "Real: \tp=%g\tL=%g\tT=%g" % (wlc.p, wlc.L, wlc.T)
    xdata = numpy.linspace(0, maxLfrac*wlc.L, npoints)
    ydata = numpy.array([-1.0]*len(xdata))
    sigma = wlc.WLC(xdata[-1])*randomScaleFactor
    for i in range(len(xdata)):
        ydata[i] = wlc.WLC(xdata[i]) + r.gauss(0, sigma)
    #for x,y in zip(xdata, ydata):
    #    print x,y
    wlc.fixedP = pfix
    wlc.fixedL = Lfix
    wlc.fixedT = Tfix
    if wlc.fixedP == False:
        wlc.p *= pgf
    if wlc.fixedL == False:
        wlc.L *= Lgf
    if wlc.fixedT == False:
        wlc.T *= Tgf
    print "Guess:\tp=%g\tL=%g\tT=%g" % (wlc.p, wlc.L, wlc.T)
    wlc.fit(xdata, ydata, textVerbose=textVerbose, plotVerbose=plotVerbose)
    print "Fit:  \tp=%g\tL=%g\tT=%g" % (wlc.p, wlc.L, wlc.T)

class AffineData (object):
    """Allow arbitaraty scaling and shifting of multichannel time series
    data (2D Numpy arrays), with access to either the raw values or
    the transformed values.
    
    >>> a = AffineData(raw_data = numpy.array([[0,1],[1,2],[3,47]]))
    >>> print a.scale
    [[1 0]
     [0 1]]
    >>> print a.shift
    [0 0]
    >>> print a.data()
    [[ 0  1]
     [ 1  2]
     [ 3 47]]
    >>> a.scale[0,0]=2
    >>> a.scale[1,1]=3
    >>> a.scale[0,1]=1
    >>> a.shift[0] = -2
    >>> a.shift[1] = -4
    >>> print a.data()
    [[ -1  -1]
     [  2   2]
     [ 51 137]]
    >>> print a.reverse([51,137])
    [  3.  47.]
    >>> print a.transform([3, 47])
    [ 51 137]
    """
    def __init__(self, raw_data, scale=None, shift=None, dtype=numpy.double):
        self.raw_data = raw_data
        assert len(self.raw_data.shape) == 2, \
            "Only supports 2D data, not %s" % self.raw_data.shape
        self.channels = self.raw_data.shape[1]
        if dtype == None:
            self.dtype = self.raw_data.dtype
        else:
            self.dtype = dtype
        if scale != None:
            self.scale = scale
        else:
            self.scale = numpy.identity(self.channels, dtype=self.dtype)
        if shift != None:
            self.shift = shift
        else:
            self.shift = numpy.zeros((self.channels,), dtype=self.dtype)
    def scaled_data(self):
        "Returns scaled data"
        return numpy.dot(self.scale, self.raw_data.transpose()).transpose()
    def data(self):
        "Returns scaled and shifted data"
        return self.scaled_data() + self.shift
    def inv_scale(self):
        """Calculates and returns the inverse of the scaling matrix.  Caches
        the scaled matrix internally, so call as often as you like ;).
        """
        calc_new_inverse = False
        if not hasattr(self, "_inv_scale_source"):
            calc_new_inverse = True
        elif (self._inv_scale_source == self.scale).all() != True:
            calc_new_inverse = True
        if calc_new_inverse == True:
            self._inv_scale_source = self.scale
            self._inv_scale = numpy.linalg.inv(self._inv_scale_source)
        return self._inv_scale
    def transform(self, vector):
        """Convert a multichannel vector from the raw coordinates to the
        transformed coordinates.
        """
        return numpy.dot(self.scale, vector) + self.shift
    def reverse(self, vector):
        """Convert a multichannel vector from the transformed coordinates back
        to raw coordinates.
        """
        return numpy.dot(self.inv_scale(), vector-self.shift)

class SawtoothModel (object):
    """Model a dataset as a series of WLCs.
    
    >>> s = SawtoothModel()
    >>> s._data = AffineData(raw_data=numpy.array([[0,1],[2,2],[1,3]]))
    >>> print s.getDataRange_matrix(1,2)
    [[1 3]
     [2 2]]
    """
    # store the data in the background so that transforms and things always work
    def _get_data(self):
        if self._real_data == None:
            self.load(override_data=numpy.array([[0,1],[0,1]]))
            fake_data = self._real_data
            self.free()
            return fake_data
        else:
            return self._real_data
    def _set_data(self, x):
        self._real_data = x
    _data = property(fget=_get_data, fset=_set_data)

    # define a few convenience methods to handle data transformations
    def _get_xshift(self):
        return self._data.shift[0]
    def _set_xshift(self, x):
        self._persist_xshift = float(x)
        self._data.shift[0] = self._persist_xshift
    xshift = property(fget=_get_xshift, fset=_set_xshift)
    
    def _get_yshift(self):
        d = self._data
        return d.shift[1]
    def _set_yshift(self, x):
        self._persist_yshift = float(x)
        self._data.shift[1] = self._persist_yshift
    yshift = property(fget=_get_yshift, fset=_set_yshift)
    
    def _get_xscale(self):
        return self._data.scale[0,0]
    def _set_xscale(self, x):
        self._persist_xscale = float(x)
        self._data.scale[0,0] = self._persist_xscale
    xscale = property(fget=_get_xscale, fset=_set_xscale)
    
    def _get_yscale(self):
        return self._data.scale[1,1]
    def _set_yscale(self, x):
        self._persist_yscale = float(x)
        self._data.scale[1,1] = self._persist_yscale
    yscale = property(fget=_get_yscale, fset=_set_yscale)
    
    def __init__(self, datafile=None, selected=True, temperature=298.15,
                 xScale=None, yScale=None, xShift=None, yShift=None,
                 persistence_length=0.4e-9, cuts=None, WLClist=None):
        self._data = None
        self._persist_xscale = xScale
        self._persist_yscale = yScale
        self._persist_xshift = xShift
        self._persist_yshift = yShift
        self.datafile = datafile
        self.selected = selected
        self.T = temperature
        self.p = persistence_length
        if cuts == None:  # cuts are the _raw_ (x,y) coordinates for slices
            self.cuts = []
        else:
            self.cuts = cuts
        if WLClist == None:
            self.WLClist = []
        else:
            self.WLClist = WLClist
    def __str__(self):
        strArr = []
        strArr.append("Datafile: %s" % self.datafile)
        if self.selected == True:
            select_value = '1'
        else:
            select_value = '0'
        strArr.append("Selected: %s" % select_value)
        strArr.append("Temperature: %f" % self.T)
        for i in range(len(self.WLClist)):
            wlc = self.WLClist[i]
            min,max = self.getMinMax(i)
            strArr.append('WLC: %g,%g,(%g,%g),(%g,%g)' % (wlc.p, wlc.L, min[0], min[1], max[0], max[1]))
        return '\n'.join(strArr)
    def __cmp__(self, other):
        "Allow sorting by datafile name"
        return cmp(self.datafile, other.datafile)
    def load(self, override_data=None):
        "Load data from a file or from an array passed in as an argument."
        if override_data == None:
            assert self.datafile != None
            data = readFile(self.datafile, split_cols=False)
        else:
            data = override_data
        self._data = AffineData(raw_data=data)
        self.fitted = False
        
        if self._persist_xscale != None:
            self.xscale = self._persist_xscale
        if self._persist_yscale != None:
            self.yscale = self._persist_yscale
        if self._persist_xshift != None:
            self.xshift = self._persist_xshift
        if self._persist_yshift != None:
            self.yshift = self._persist_yshift
    def free(self):
        "Release the data to free memory"
        del(self._real_data)
        self._data = None
        self.fitted = False
    
    # define a few more convenience methods to handle data transformations
    def transform(self, vector):
        return self._data.transform(vector)
    def reverse(self, vector):
        return self._data.reverse(vector)
    
    # define a few convenience methods to handle data access
    def xdata(self):
        return self._data.data()[:,0]
    def raw_xdata(self):
        return self._data.raw_data[:,0]
    def ydata(self):
        return self._data.data()[:,1]
    def raw_ydata(self):
        return self._data.raw_data[:,1]
    def data(self):
        data = self._data.data()
        return (data[:,0], data[:,1])
    def raw_data(self):
        return self._data.raw_data
    
    def getDataRange_matrix(self, xmin=None, xmax=None):
        """Return data where xmin <= x <= xmax.  xmin, xmax, and the returned
        data are in the scaled/shifted coordinates.
        """
        if xmin != None and xmax != None:
            assert xmin <= xmax
        data = self._data.data()
        data = data[data[:,0].argsort(),]

        imin = numpy.searchsorted(data[:,0], xmin, side='left')
        imax = numpy.searchsorted(data[:,0], xmax, side='right')

        return data[imin:imax,:]

    def getDataRange(self, xmin=None, xmax=None):
        """Return the output of getDataRange_matrix split into x/y columns"""
        data = self.getDataRange_matrix(xmin=xmin, xmax=xmax)
        return (data[:,0], data[:,1])
    
    # fitting methods
    def getMinMax_raw(self, WLC_index):
        if len(self.cuts) < 2*(WLC_index+1):
            raise IndexError, "WLC_index %d to big for number of cuts %d" % (WLC_index, len(self.cuts))
        if type(self.cuts) == list:
            self.cuts = numpy.array(self.cuts)
        self.cuts = self.cuts[self.cuts[:,0].argsort(),]
        if self.xscale < 0:
            self.cuts = self.cuts[::-1] # reverse array
        min = self.cuts[2*WLC_index]
        max = self.cuts[2*WLC_index+1]
        return (min, max)
    
    def getMinMax(self, WLC_index):
        min_r,max_r = self.getMinMax_raw(WLC_index)
        min = self.transform(min_r)
        max = self.transform(max_r)
        return (min, max)
    
    def getDataForWLC(self, WLC_index):
        min,max = self.getMinMax(WLC_index)
        xdata, ydata = self.getDataRange(min[0], max[0])
        return (xdata, ydata)
    
    def zero_y(self, technique=None, npoints=500):
        if technique == None:
            if len(self.cuts) > 0:
                technique = "from last cut"
            else:
                technique = "from tail"
        xdata, ydata = self.data()
        if technique == "from contact":
            # Average for all points off of the surface
            istart = numpy.searchsorted(xdata, 0, side='left')
            y_avg_data = ydata[istart:]
        elif technique == "from tail":
            y_avg_data = ydata[-npoints:]
        elif technique == "from last cut":
            xcut = self.transform(self.cuts[-1])[0]
            istart = numpy.searchsorted(xdata, xcut, side='left')
            y_avg_data = ydata[istart:istart+npoints]
        else:
            assert 1==0, "Unknown technique '%s'" % technique
        y_avg = float(sum(y_avg_data)) / len(y_avg_data)
        self.yshift -= y_avg

    def fit(self, textVerbose=False, plotVerbose=False):
        """Fit WLCs to data between cut pairs.
        """
        WLClist = []
        for i in range(len(self.cuts)/2):
            min,max = self.getMinMax(i)
            if textVerbose == True:
                print "Fitting the %dth WLC (%s to $s)" % (i, min, max)
            xdata,ydata = self.getDataForWLC(i)
            # TODO: get info from previous WLC fit?
            wlc = WLC(p=self.p, L=2*max[0], T=self.T,
                      fixedP=True, fixedL=False, fixedT=True)
            wlc.fit(xdata, ydata, textVerbose=textVerbose,
                    plotVerbose=plotVerbose)
            WLClist.append(wlc)
        self.WLClist = WLClist
        self.fitted = True

    def genFitDataForWLC(self, WLC_index):
        assert WLC_index < len(self.WLClist)
        xdata,ydata = self.getDataForWLC(WLC_index)
        wlc = self.WLClist[WLC_index]
        ymodel = wlc.WLC(xdata)
        return (xdata, ymodel)
    
    def genFitData(self):
        """Generate x and y data arrays for plotting the WLC fits.
        """
        if self.fitted == False:
            return (None, None)
        fullXdata = []
        fullYdata = []
        for i in range(len(self.cuts)/2):
            try:
                xdata,ymodel = self.genFitDataForWLC(i)
            except AssertionError, e:
                print str(e)
                break
            fullXdata.extend(xdata)
            fullYdata.extend(ymodel)
        return (fullXdata, fullYdata)

    # unfolding location extraction
    def getUnfolds(self, return_hulls=False):
        """Extract unfolding (location,force) pairs.  Currently defined as the
        intersection between the WLC fit and the convex hull
        surrounding the associated cut data.  The intersection with
        the largest X value is chosen (there will be either 2 or 4
        intersection points).  Currently the crossings are found by
        brute force comparing (xdata,ymodel) pairs with the convex
        hull.
        """
        unfolds = []
        hulls = []
        for i in range(len(self.WLClist)):
            xdata,ydata = self.getDataForWLC(i)
            xdata,ymodel = self.genFitDataForWLC(i)
            data = numpy.array([[x,y] for x,y in zip(xdata, ydata)])
            modelpoints = numpy.array([[x,y] for x,y in zip(xdata, ymodel)])
            hull = quickhull2d.qhull(data)
            inside = quickhull2d.points_inside_hull(hull, modelpoints)
            unfold = None
            for i in range(len(ymodel)-1, 0, -1):
                if inside[i] == True:
                    unfold = modelpoints[i]
                    break
            unfolds.append(unfold)
            hulls.append(hull)
        if return_hulls == True:
            return (unfolds, hulls)
        else:
            return unfolds
    # Works in progress
    def peak(self) :
        """Shift the Xmaxs so that they lie on the peak unfolding force measurement"""
        raise NotImplementedError
        self.fit()
        if len(self.model.WLClist) == 0 :
            return
        xdata = self.model.xdata()
        ydata = self.model.ydata()
        for i in range(len(self.model.WLClist)) :
            xWLCmax = self.model.Xcuts[2*i+1] + self.model.xShift
            raise NotImplementedError, "New mark manager format"
            imin,Xmin = self.markManager.snapToData(xWLCmax, xdata, ydata)
            Ymin = ydata[imin]*self.peakThreshold
            imax = imin
            Ypeak = ydata[imin]
            ipeak = imin
            while imax < len(ydata) and ydata[imax] >= Ymin :
                if ydata[imax] >= Ypeak :
                    Ypeak = ydata[imax]
                    ipeak = imax
                imax += 1
            #print "peak %d moved from (%g,%g) to (%g,%g)" % (i, Xmin*1e9, Ymin*1e12, xdata[ipeak]*1e9, Ypeak*1e12)
            self.model.Xcuts[2*i+1] = xdata[ipeak] - self.model.xShift
        self.fit()

    

def _testSawtoothModel(p=0.4e-9, L=28, T=298.15, numWLCs=4,
                       pgf=0.367879, Lgf=3.14159, Tgf=1.0,
                       pfix=True, Lfix=False, Tfix=True, randomScaleFactor=0.1,
                       npoints=300, maxLfrac=0.9,
                       textVerbose=False, plotVerbose=False):
    """Xgf is the "Guess Factor" for parameter x.  For example, pGuess = p*pgf
    The Guess Factor is only applied to fitted (unfixed) parameters.
    """
    print "Testing SawtoothModel"    
    from random import Random
    assert Lgf > maxLfrac, "Evaluation would have some x > L"
    r = Random()
    def makeWLC(xstart, xstop, p=p, L=L, T=T):
        wlc = WLC(p=p, L=L, T=T)
        xdata = numpy.linspace(xstart, xstop, npoints)
        cuts = [[xstart + 0.1*(xstop-xstart), 0], [xstop, 0]]
        ydata = numpy.array([-1.0]*npoints)
        sigma = wlc.WLC(xstop)*randomScaleFactor
        for i in range(len(xdata)):
            ydata[i] = wlc.WLC(xdata[i]) + r.gauss(0, sigma)
        unfold = (xstop, wlc.WLC(xstop))
        return (xdata, ydata, cuts, unfold)

    xdata,ydata,cuts,unfold = makeWLC(xstart=0, xstop=maxLfrac*L, L=L)
    unfolds = [unfold]
    for i in range(1, numWLCs):
        x,y,c,u = makeWLC(xstart=1.01*xdata[-1], xstop=maxLfrac*L+xdata[-1],
                           L=(i+1)*L)
        xdata = numpy.concatenate([xdata, x])
        ydata = numpy.concatenate([ydata, y])
        cuts = numpy.concatenate([cuts, c])
        unfolds.append(u)
    
    s = SawtoothModel(temperature=T, persistence_length=p, cuts=cuts)
    s._data = AffineData(raw_data=numpy.array([xdata,ydata]).transpose())
    s.fit()
    xfit,yfit = s.genFitData()
    print "Real unfolds:"
    for point in unfolds:
        print "  %.3g\t%.3g" % point
    print "Fit unfolds:"
    fitUnfolds,hulls = s.getUnfolds(return_hulls=True)
    for point in fitUnfolds:
        print "  %.3g\t%.3g" % tuple(point)
    xunfold = [p[0] for p in fitUnfolds]
    yunfold = [p[1] for p in fitUnfolds]
    
    if plotVerbose == True:
        import pylab
        pylab.figure()
        pylab.hold(True)
        for hull in hulls:
            xhull = [p[0] for p in hull]
            yhull = [p[1] for p in hull]    
            pylab.plot(xhull, yhull, 'b:')
        pylab.plot(xunfold, yunfold, 'gx',
                   xdata, ydata, 'k.', xfit, yfit, 'r-')
        pylab.show()

def _test():
    import doctest
    doctest.testmod()
    _testWLC(plotVerbose=False)
    _testSawtoothModel(plotVerbose=True)

if __name__ == "__main__":
    _test()
