#!/usr/bin/python
#
# sawmodel_mark - GUI interface for marking sawtooth WLCs in force
# spectroscopy unfolding data.
#
# Copyright (C) 2008-2009, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.
"""Provide a graphical interface for fitting WLCs to unfolding force
curves and extracting unfolding forces."""

import Tkinter as tk
import matplotlib
matplotlib.use("TKAgg")
#matplotlib.rc('font', **{'family': 'sans-serif',
#             'sans-serif':['Helvitica', 'Arial']})
## I was getting a 'could not match ...' error fromvfont_manager.py:989
# searched for available fonts on my system with
#>>> from matplotlib.font_manager import *
#>>> p = FontProperties()
#>>> p.get_family()
#>>> p.get_name()
#matplotlib.rc('font', **{'family': 'serif', 'serif':['cmtt10']})
from pylab import plot, hold, draw, show, connect
from matplotlib.backends.backend_tkagg \
    import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure
from numpy import array, sqrt
from scipy.optimize import leastsq
import os.path

import sawmodel.fit_xml as fit_xml

class Point (object) :
    """
    A 2-D vector class
    
    >>> o = Point(0,0)
    >>> p = Point(3,4)
    >>> print p.dist(o)
    5.0
    >>> Point.yscale = 1e100
    >>> print p.dist(o)
    3.0
    """
    xscale = 1
    yscale = 1
    def __init__(self, x, y) :
        self.x = float(x)
        self.y = float(y)
    def __str__(self) :
        return "(%s, %s)" % (self.x, self.y)
    def __repr__(self) :
        return "(%s, %s)" % (repr(self.x), repr(self.y))        
    def __cmp__(self, other) :
        "Sort first by x, then by y"
        if self.x != other.x :
            return cmp(self.x, other.x)
        elif self.y != other.y :
            return cmp(self.y, other.y)
        return 0
    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)
    def mag(self):
        return sqrt((self.x/Point.xscale)**2 + (self.y/Point.yscale)**2)
    def dist(self, other):
        return (self-other).mag()

class MarkManager (object) :
    """
    Allow easy point addition/removal from an array
    
    >>> m = MarkManager()
    >>> m.addMark(3,4)
    >>> print m
    [(3.0, 4.0)]
    >>> m.removeMark(3,4)
    >>> print m
    []
    >>> m.snapToData(1, 5, [0,1,2,3,4], [-1,2,5,8,11])
    (2, 3.0, 8.0)
    """
    def __init__(self, marks=None) :
        if marks == None :
            self.marks = []
        else :
            self.marks = marks
    def __repr__(self) :
        return repr(self.marks)
    def xShift(self) :
        raise NotImplementedError
    def yShift(self) :
        raise NotImplementedError
    def addMark(self, x, y) :
        p = Point(x, y)
        print "adding", p
        #assert p not in self.marks, "%s\n%s" % (p, self)
        if p in self.marks:
            print "Dropping duplicate mark %s" % (p)
            return
        self.marks.append(p)
        self.marks.sort()
    def removeMark(self, x, y) :
        p = Point(x, y)
        try :
            self.marks.remove(p)
        except ValueError :
            pass # don't worry if x wasn't in our list...
    def snapToData(self, x, y, xdata, ydata) :
        """
        Return (index, xvalue, yvalue) for the datapoint nearest (x,y)
        """
        if len(xdata) == 0:
            return (None, x, y)
        
        p = Point(x,y)
        #print "snapping to", p, "from"
        #for x,y in zip(xdata,ydata):
        #    print '\t',x,'\t',y
        
        data = []
        for xd,yd in zip(xdata, ydata) :
            data.append(Point(xd,yd))
        data.sort()
        
        # find the point with a nearly matching x value
        i = 0
        while True: # could be improved: binomial search
            if i >= len(data) :
                break
            elif data[i].x >= p.x :
                break
            i += 1
        if i >= len(data) : # don't go past last data point
            i = -1
        #print "nearest >= x point is %d %s" % (i, data[i])
        
        # grab all points with x values that 'might be' 
        r = p.dist(data[i]) # distance to a point with similar x
        close_data = [pd for pd in data if abs(p.x-pd.x)/Point.xscale <= r]
        pmin = data[i]
        for pd in close_data :
            r_try = p.dist(pd) 
            if r_try < r :
                r = r_try
                pmin = pd
        for i,x,y in zip(range(len(xdata)), xdata, ydata):
            if pmin.x == x and pmin.y == y:
                imin = i
        
        if False:
            # check closest point algorithm
            Figure()
            xd = [pd.x for pd in close_data]
            yd = [pd.y for pd in close_data]
            plot(xd, yd, 'r.')
            hold(True)
            plot([p.x, pmin.x], [p.y, pmin.y], 'bo')
            show()
        
        return (imin, pmin.x, pmin.y)
    def nearestMark(self, x, y) :
        return self.snapToData(x, y,
                               [p.x for p in self.marks],
                               [p.y for p in self.marks])
    def markFromX(self, x):
        p = [p for p in self.marks if p.x == x]
        return (p[0].x, p[0].y)
    def initializeFromCutsAndData(self, cuts, xdata, ydata) :
        self.marks = []
        i = -1
        if cuts == None:
            return
        for c in cuts:
            i,x,y = self.snapToData(c[0], c[1], xdata, ydata)
            self.addMark(x, y)

class FitSelector (object) :
    """Factory for fitting unfolding curves"""
    def __init__(self, model, p=0.4e-9) :
        self.model = model
        self.p = p
        self.T = self.model.T
        self.markManager = MarkManager()
    def load(self):
        self.model.load()
        print "loading", self.model.datafile
        self.markManager.initializeFromCutsAndData(self.model.cuts,
                                                   self.model.raw_xdata(),
                                                   self.model.raw_ydata())
        self.fitted = False
        self.peakThreshold = 0.9 # fraction falloff in force before we give up on finding a nicer peak
    def free(self) :
        self.model.free()
    def __str__(self) :
        return "<Fit Selector '%s', fitted %s to %d marks>" % (self.model.datafile, self.fitted, len(self.model.Xcuts))
    def cleanMarks(self) :
        pass
    def fit(self) :
        self.model.fit()
        self.fitted = self.model.fitted

class FitSelectorGUI (object) :
    """Provide a graphical interface for fitting WLCs to the data."""
    def __init__(self, parent, unfoldModels, saveFileName=None, start_index=0) :
        self.parent = parent
        self.frame = tk.Frame(self.parent)
        self.frame.pack(anchor='nw', expand=tk.YES, fill=tk.BOTH)

        # setup figure and axes
        self.figure = Figure(figsize=(5,4), dpi=50)
        self.axes = self.figure.add_subplot(111)
        
        self.frameFigure = tk.Frame(self.frame)
        self.frameFigure.pack(side=tk.TOP, anchor='nw', expand=tk.YES, fill=tk.BOTH)
        self.frameToolbar = tk.Frame(self.frame)
        self.frameToolbar.pack(anchor='nw')
        self.frameNonplot = tk.Frame(self.frame)
        self.frameNonplot.pack(anchor='nw')
        self.frameInput = tk.Frame(self.frameNonplot)
        self.frameInput.pack(side=tk.LEFT, anchor='nw')
        self.frameOutput = tk.Frame(self.frameNonplot)
        self.frameOutput.pack(anchor='nw')
        
        self.canvas = FigureCanvasTkAgg(self.figure, master=self.frameFigure)
        self.canvas.show()
        self.canvas.get_tk_widget().pack(fill=tk.BOTH, expand=tk.YES)
        #self.parent.winfo_toplevel().update_idletasks() # from James Stroud Thu, 26 Apr 2007 21:24:52 -0700 http://coding.derkeiler.com/Archive/Python/comp.lang.python/2007-04/msg04162.html
        self.toolbar = NavigationToolbar2TkAgg(self.canvas, self.frameToolbar)
        self.toolbar.update()
        
        # add input widgets and callbacks
        self._addButton(self.frameInput, "markButton", "Toggle mark mode", self.mark)
        self._addButton(self.frameInput, "shiftButton", "Toggle shift mode", self.shift)
        self._addButton(self.frameInput, "zero_yButton", "Zero Y", self.zero_y)
        self._addButton(self.frameInput, "fitButton", "Fit", self.fit)
        #self._addButton(self.frameInput, "peakButton", "Tweak Xmaxs to peaks", self.peak)
        self._addCheckbox(self.frameInput, "selectBox", "Good fit", self.select)
        self._addButton(self.frameInput, "nextButton", "Next Curve", self.next)
        self._addButton(self.frameInput, "prevButton", "Previous Curve", self.prev)
        self._addButton(self.frameInput, "saveButton", "Save All", self.save)
        self._addButton(self.frameInput, "quitButton", "Quit", self.quit)
        self.mark(None) # start out in mark mode
        
        # add output widgets and callbacks
        self._addLabeledTextout(self.frameOutput, "TLabel", "Temperature T (C)")
        self._addLabeledTextout(self.frameOutput, "pLabel", "Persistance length p (nm)")
        self._addLabeledTextout(self.frameOutput, "LLabel", "Contour length L (nm)")
        self._addLabeledTextout(self.frameOutput, "dLLabel", "Delta L (nm)")

        # setup plots for data, marks, and fits
        self.zeroLinestyle = 'g-'
        self.dataLinestyle = 'k.'
        self.markLinestyle = 'ro'
        self.fitLinestyle = 'b.'
        self.axes.hold(True)
        self.axes.set_xlabel('Extension (m)')
        self.axes.set_ylabel('Force (N)')
        self.zeroPlots = self.axes.plot([0,1],[0,0],self.zeroLinestyle,
                                        [0,0],[0,1], self.zeroLinestyle) # dummy data
        self.dataPlot = self.axes.plot([0,1],[0,0],self.dataLinestyle)[0] # dummy data
        self.fitPlot = self.axes.plot([0,1],[0,0],self.fitLinestyle)[0] # dummy data
        self.markPlot = self.axes.plot([0,1],[0,0],self.markLinestyle)[0] # dummy data
        
        # setup list of unfolding datasets
        self.unfoldModels = unfoldModels
        self.saveFileName = saveFileName
        self.index = start_index
        self.selector = None
        self.loadCurrent()
    def _addButton(self, frame, attrname, text, callback) :
        button = tk.Button(frame)
        button.configure(text=text, width=20)
        button.pack(anchor='nw', side=tk.TOP)
        button.bind("<Button-1>", callback)
        button.bind("<Return>", callback)
        setattr(self, attrname, button)
    def _addLabeledTextout(self, frame, attrname, text) :
        labelFrame = tk.Frame(frame)
        labelFrame.pack(anchor='nw', side=tk.TOP)
        labelLabel = tk.Label(labelFrame, text=text, width=30)
        labelLabel.pack(anchor='nw', side=tk.LEFT)
        labelText = tk.Label(labelFrame, text='-')
        labelText.pack(anchor='nw')
        setattr(self, 'labeledText'+attrname, {"frame":labelFrame, "label":labelLabel, "text":labelText})
        setattr(self, 'set_'+attrname, labelText.configure)
    def _addCheckbox(self, frame, attrname, text, callback) :
        var = tk.IntVar()
        check = tk.Checkbutton(frame, text=text, variable=var, command=callback)
        check.pack(anchor='nw', side=tk.TOP)
        check.var = var
        setattr(self, attrname, check)
    def loadCurrent(self) :
        if self.selector != None :
            self.selector.free()
        self.selector = FitSelector(self.unfoldModels[self.index])
        self.selector.load()
        self.setSelect(self.selector.model.selected)
        self.set_pLabel(text=str(self.selector.p*1e9))    # display p in nm
        self.set_TLabel(text=str(self.selector.T-273.15)) # display temp in Celsius
        title = os.path.basename(self.selector.model.datafile)
        self.axes.set_title(title)
        if len(self.selector.model.cuts) >= 2 :
            self.selector.fitted = True
            self.markPlot.set_visible(True)
            self.fitPlot.set_visible(True)
        else :
            self.markPlot.set_visible(False)
            self.fitPlot.set_visible(False)
        self.setLLabelText()
        self.setAxisTight()
        self.draw()
    def setLLabelText(self) :
        if len(self.selector.model.cuts) >= 2 :
            strArr = []
            Ls = []
            for i in range(len(self.selector.model.cuts)/2) :
                if i < (self.selector.model.WLClist) :
                    L = self.selector.model.WLClist[i].L
                    Ls.append(L)
                else :
                    L = '-'
                min,max = self.selector.model.getMinMax(i)
                u = abs(self.selector.model.xscale)
                strArr.append("%g (%g, %g)" % (L/u, min[0]/u, max[0]/u)) # convert L to raw data units
            self.set_LLabel(text='\n'.join(strArr))
            if len(Ls) > 1 :
                dLstr = [str((Lb-La)/u) for Lb,La in zip(Ls[1:],Ls)] # convert dL to raw data units
            else :
                dLstr = ['-']
            self.set_dLLabel(text='\n'.join(dLstr))
        else :
            self.set_LLabel(text='-')
            self.set_dLLabel(text='-')
    def setAxisRange(self, xmin, xmax, ymin, ymax) :
        self.axes.axis([xmin,xmax,ymin,ymax])
        self.zeroPlots[0].set_data([xmin,xmax],[0,0])
        self.zeroPlots[1].set_data([0,0],[ymin,ymax])
    def setAxisTight(self) :
        xdata = self.selector.model.xdata()
        ydata = self.selector.model.ydata()
        xmin = min(xdata)
        xmax = max(xdata)
        ymin = min(ydata)
        ymax = max(ydata)
        # Protein unfolding specific window size
        #ymin = min(self.selector.model.ydata) - 10 # pN
        #assert ymin < 0
        #ymax = ymin * -0.5
        #xmax = 50
        #xmin = -500 # nm
        #self.subplot.autoscale_view(tight=True) # not working for some reason...
        self.setAxisRange(xmin, xmax, ymin, ymax)
    def shiftAxisRange(self, dx, dy) :
        xmin,xmax,ymin,ymax = self.axes.axis()
        xmin += dx
        xmax += dx
        ymin += dy
        ymax += dy
        self.setAxisRange(xmin, xmax, ymin, ymax)
    def draw(self) :
        self.dataPlot.set_data(self.selector.model.xdata(), self.selector.model.ydata())
        if len(self.selector.markManager.marks) > 0 :
            self.markPlot.set_visible(True)
# WORKING: keep Xcuts up to date...
            #xdata = self.selector.model.xdata()
            #ydata = self.selector.model.ydata()
            #xmark = [x+self.selector.model.xShift for x in self.selector.model.Xcuts]
            #ymark = []
            #for x in self.selector.model.Xcuts :
            #    x,y = self.selector.markManager.markFromX(x)
            #    ymark.append(y+self.selector.model.yShift)
            marks = [self.selector.model.transform([p.x,p.y])
                     for p in self.selector.markManager.marks]
            xmark = [m[0] for m in marks]
            ymark = [m[1] for m in marks]
            self.markPlot.set_data(xmark, ymark)
            if self.selector.fitted == True :
                self.fitPlot.set_visible(True)
                xfit, yfit = self.selector.model.genFitData()
                self.fitPlot.set_data(xfit, yfit)
            else :
                self.fitPlot.set_visible(False)
        else :
            self.markPlot.set_visible(False)
            self.fitPlot.set_visible(False)
        self.canvas.show()
    def setSelect(self, value) :
        self.selectBox.var.set(value)
    def select(self) :
        val = self.selectBox.var.get()
        if val == 1 :
            self.selector.model.selected = True
        else :
            self.selector.model.selected = False
    def next(self, event) :
        if self.index+1 < len(self.unfoldModels) :
            self.index += 1
            self.loadCurrent()
    def prev(self, event) :
        if self.index > 0 :
            self.index -= 1
            self.loadCurrent()
    def zero_y(self, event=None):
        self.selector.model.zero_y()
        self.fit()
    def fit(self, event=None) :
        self.selector.fit()
        self.setLLabelText()
        self.draw()
    def peak(self, event) :
        self.selector.peak()
        self.setLLabelText()
        self.draw()
    def save(self, event) :
        fit_xml.saveUnfoldList(self.unfoldModels, self.saveFileName)
        print self.saveFileName, "saved"
    def toolbarButton(self, name, activeFlag, event) :
        'Activate/deactivate a toolbar tool.  See matplotib/backend_bases.py NavigationToolbar2.pan()'
        tb = self.toolbar
        if tb._active == activeFlag :
            tb._active = None
        else :
            tb._active = activeFlag
        callbacktypes = ['press', 'release', 'drag']
        callbackstrings = {'press':'button_press_event',
                           'release':'button_release_event',
                           'drag':'motion_notify_event'}
        for callbacktype in callbacktypes :
            attrname = '_id%s' % callbacktype.capitalize() # eg: "_idPress"
            id = getattr(tb, attrname)
            if id is not None :
                setattr(tb, attrname, tb.canvas.mpl_disconnect(id))
                tb.mode = ''
        if tb._active :
            for callbacktype in callbacktypes :
                idattrname = '_id%s' % callbacktype.capitalize() # e.g. "_idPress"
                callbackfn = '%s_%s' % (callbacktype, name) # e.g. "press_zoom"
                if hasattr(self, callbackfn) :
                    callbackstring = callbackstrings[callbacktype] # e.g. "button_press_event"
                    setattr(tb, idattrname, tb.canvas.mpl_connect(callbackstring, getattr(self, callbackfn)))
            tb.mode = '%s mode' % name
            tb.canvas.widgetlock(tb)
        else :
            tb.canvas.widgetlock.release(tb)
        for a in tb.canvas.figure.get_axes() :
            a.set_navigate_mode(tb._active)
        tb.set_message(tb.mode)
    def mark(self, event) :
        'Activate/deactivate the mark tool.  Mark with left button, unmark with right.'
        self.toolbarButton('mark', 'MARK', event)
    def release_mark(self, event) :
        'the release mouse button callback in mark mode.  See matplotlib/backend_bases.py'
        if event.inaxes == self.axes :
            if event.button == 1 : # left click
                self.addMark(event.xdata, event.ydata)
            elif event.button == 3 : # right click
                self.removeMark(event.xdata, event.ydata)
    def addMark(self, x, y) :
        # Correct the scaling based on window squareness.
        xmin,xmax,ymin,ymax = self.axes.axis()
        Point.xscale = (xmax-xmin)
        Point.yscale = (ymax-ymin)
        # snap to transformed data
        i,xsnap,ysnap = \
            self.selector.markManager.snapToData(x, y,
                                                 self.selector.model.xdata(),
                                                 self.selector.model.ydata())
        # add mark from raw data
        self.selector.markManager.addMark(self.selector.model.raw_xdata()[i],
                                          self.selector.model.raw_ydata()[i])
        self.selector.model.cuts = [(p.x,p.y) for p in self.selector.markManager.marks]
        self.draw()
    def removeMark(self, x, y) :
        # Correct the scaling based on window squareness and data
        # scaling.  This way we get proper point-point distances after
        # converting back into raw coordinates.
        xmin,xmax,ymin,ymax = self.axes.axis()
        Point.xscale = (xmax-xmin)/abs(self.selector.model.xscale)
        Point.yscale = (ymax-ymin)/abs(self.selector.model.yscale)
        # convert to raw coords
        vect = self.selector.model.reverse([x,y])
        # snap in raw coords
        i,xsnap,ysnap = \
            self.selector.markManager.nearestMark(vect[0], vect[1])
        self.selector.markManager.removeMark(xsnap, ysnap)
        self.selector.model.cuts = [(p.x,p.y) for p in self.selector.markManager.marks]
        self.draw()
    def shift(self, event) :
        'Activate/deactivate the shift tool.  Shift by dragging with left button.'
        self._shift_active=False
        self.toolbarButton('shift', 'SHIFT', event)
    def press_shift(self, event) :
        if event.inaxes == self.axes and event.button == 1 :
            self._shift_active = True
            self._shift_x = event.xdata
            self._shift_y = event.ydata
            self._shift_xShift_orig = self.selector.model.xshift
            self._shift_yShift_orig = self.selector.model.yshift
    def release_shift(self, event) :
        if event.button == 1 :
            if self._shift_active == True :
                dx = event.xdata - self._shift_x
                dy = event.ydata - self._shift_y
                self.shiftAxisRange(dx,dy)
                self._shift_x = None
                self._shift_y = None
                self.draw()
            self._shift_active = False
    def drag_shift(self, event) :
        'the release mouse button callback in mark mode.  See matplotlib/backend_bases.py'
        if event.inaxes == self.axes and self._shift_active :
            dx = event.xdata - self._shift_x
            dy = event.ydata - self._shift_y
            self.selector.model.xshift = self._shift_xShift_orig + dx
            self.selector.model.yshift = self._shift_yShift_orig + dy
            self.fit()
            self.draw()
    def quit(self, event) :
        self.parent.destroy()

def _test() :
    import doctest
    doctest.testmod()
    #_testWLC(plotVerbose=True)

def main():
    import sys
    
    if len(sys.argv) <= 1 :
        _test()
    else :
        listfilename = sys.argv[1]
        if len(sys.argv) >= 3:
            startfile = sys.argv[2]
        else:
            startfile = None
        
        unfoldModels = fit_xml.readUnfoldList(listfilename)
        if startfile != None:
            files = [m.datafile for m in unfoldModels]
            index = files.index(startfile)
            print "start index", index
        else:
            index = 0
        
        root = tk.Tk()
        gui = FitSelectorGUI(root, unfoldModels=unfoldModels, saveFileName=listfilename, start_index=index)
        root.mainloop()
    return 0

if __name__ == "__main__" :
    main()
