#!/usr/bin/python
#
# sawmodel_select - GUI interface for selecting "good" force
# spectroscopy unfolding curves.
#
# Copyright (C) 2008-2009, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

"""Provide a graphical interface for browsing and selecting unfolding
force curves."""

import matplotlib
#matplotlib.use("TKAgg")
matplotlib.use("GTKAgg")
import pylab
import matplotlib.widgets as widget
from numpy import array, fromfile, double
import os.path
import sys

VERBOSE = 3

class Subplot (object) :
    def __init__(self, plotNum, numRows, numCols, figure=None) :
        self.plotNum = plotNum
        self.numRows = numRows
        self.numCols = numCols
        self.figure = figure
        self.data = None
        self.title = None
        self.selected = False
        self.selectCallback = None
        self.guiSetup()
    def guiSetup(self) :
        "GUI specific"
        if self.figure == None :
            self.figure = pylab.figure()
        self.subplot = self.figure.add_subplot(self.numRows, self.numCols, self.plotNum)
        pylab.connect('button_press_event', self.toggleSelection)
    def setData(self, xdata, ydata, title="", selected=False, xrange=None, yrange=None) :
        "GUI specific"
        if self.data == None :
            self.plot = self.subplot.plot(xdata, ydata)
        else :
            self.plot[0].set_data(xdata, ydata)
        self.data = (xdata, ydata)
        self.title = title
        self.selected = selected
        self.setAxisRange()
        self.setSelection(withCallback=False)
    def setAxisRange(self) :
        #xmin = min(self.data[0])
        #xmax = max(self.data[0])
        ymin = min(self.data[1])
        #ymax = max(self.data[1])
        # Protein unfolding specific window size
        assert ymin < 0
        ymax = ymin * -0.5
        xmax = 0
        xmin = -300 # nm
        self.subplot.axis([xmin,xmax,ymin,ymax])
        #self.subplot.autoscale_view(tight=True) # not working for some reason...
    def setSelection(self, withCallback=True) :
        if self.selected :
            bg = 'white'
        else :
            bg = 'grey'
        self.subplot.set_title(self.title, backgroundcolor=bg)
        if self.selectCallback and withCallback :
            self.selectCallback(self.selected)
    def toggleSelection(self, event) :
        if event.inaxes == self.subplot :
            self.selected = not self.selected
            self.setSelection()
            pylab.draw()

class MultiFigure (object) :
    def __init__(self, rows=3, cols=3, reservedAxes=0) :
        self.figure = pylab.figure()
        self.rows = rows
        self.cols = cols
        self.nPlots = self.rows*self.cols - reservedAxes
        self.subplots = []
        for row in range(self.rows) :
            for col in range(self.cols) :
                plotNum = 1 + row + col*self.rows
                self.subplots.append(Subplot(plotNum, self.rows, self.cols, self.figure))
    def newData(self, newdata, titles=None, selects=None) :
        assert len(newdata) <= self.nPlots
        title=""
        select = False
        for i in range(self.nPlots) :
            if i < len(newdata) :
                data = newdata[i]
                if titles != None :
                    title = titles[i]
                if selects != None :
                    select = selects[i]
                self.subplots[i].setData(data[0], data[1], title, select)
                self.subplots[i].visible = True
            else :
                self.subplots[i].visible = False
        pylab.draw()

def readFile(fid, cols=2, sep='\t'):
    """This is the lower level reader.  It works on all file types, but
    you need to know the number of columns in the file ahead of time."""
    data = fromfile(file=fid, dtype=double, sep=sep)
    rows = len(data)/cols
    data = data.reshape((rows,cols))
    x = data[:,0]
    y = data[:,1]
    return x,y

class SelectCallback (object) :
    def __init__(self, name, fn) :
        self.name = name
        self.fn = fn
    def __call__(self, selected) :
        self.fn(self.name, selected)

def readFileSelections(filename) :
    fileList = {}
    for line in file(filename, 'r') :
        line = line.split('\n')[0] # strip trailing newline
        if len(line) == 0 or line[0] == '#' :
            continue # blank or comment
        fields = line.split()
        name = fields[0]
        if len(fields) == 2 :
            selectString = fields[1]
            if selectString == '1' :
                selected = True
            else :
                selected = False
            fileList[name] = selected
        elif len(fields) == 1 :
            fileList[name] = False
        else :
            raise Exception, "Invalid line in fileList %s:\n%s" % (filename, line)
    return fileList

def saveFileSelections(fileList, filename=None) :
    keys = fileList.keys()
    keys.sort()

    if filename != None :
        fid = file(filename, 'w')
        if VERBOSE >= 1:
            print >> sys.stderr, "Saving to %s" % filename
    else :
        fid = sys.stdout

    for key in keys :
        if fileList[key] == True :
            selected = 1
        else :
            selected = 0
        print >> fid, "%s\t%d" % (key, selected)

    if filename != None :
        fid.close()
    if VERBOSE >= 1:
        print >> sys.stderr, "Saved"

class UnfoldSelector (MultiFigure) :
    def __init__(self, selectedFiles, saveFileName=None, **kwargs) :
        MultiFigure.__init__(self, reservedAxes=1, **kwargs)
        self.makeWidgetAxes(4)
        self.nextButton = widget.Button(self.widgetAxes[0], "Next batch")
        self.nextButton.on_clicked(self.next)
        self.prevButton = widget.Button(self.widgetAxes[1], "Previous batch")
        self.prevButton.on_clicked(self.prev)
        self.viewRadio = widget.RadioButtons(self.widgetAxes[2], ["All", "Selected", "Not selected"])
        self.viewRadio.on_clicked(self.setView)
        self.saveButton = widget.Button(self.widgetAxes[3], "Save selections")
        self.saveButton.on_clicked(self.save)
        for i in range(self.nPlots) :
            sp = self.subplots[i]
            sp.selectCallback = SelectCallback(i, self.changeSelection)
        self.selectedFiles = selectedFiles # array tracking file selections
        self.fileList = []
        for file,active in self.selectedFiles.items() :
            self.fileList.append(file)
        self.fileList.sort()
        self.saveFileName = saveFileName
        self.view = "All"
        self.index = 0
        self.showData()
    def makeWidgetAxes(self, numWidgets) :
        widgetAxes = self.subplots[-1].subplot.axes
        widgetAxes.set_visible(False)
        left = widgetAxes.figLeft
        bot = widgetAxes.figBottom
        w = widgetAxes.figW
        h = widgetAxes.figH
        self.widgetAxes = []
        dh = float(h)/numWidgets
        for i in range(numWidgets) :
            self.widgetAxes.append(self.figure.add_axes([left, bot+h-(i+1)*dh,w,dh]))
    def showData(self) :
        "a higher level wrapper around newData()"
        files = self._getCurrentFiles()
        data = []
        titles = []
        selects = []
        for filename in files :
            if filename != None :
                xdata,ydata = readFile(filename, cols=2)
                titles.append(os.path.basename(filename))
                selects.append(self.selectedFiles[filename])
            else :
                break
            data.append((xdata,ydata))
        self.newData(data, titles, selects)
    def _isValidIndex(self, index) :
        name = self.fileList[index]
        if self.view == "All" :
            return True
        elif self.view == "Selected" and self.selectedFiles[name] == True :
            return True
        elif self.view == "Not selected" and self.selectedFiles[name] == False :
            return True
        return False
    def _getValidIndex(self, startIndex) :
        "will return startIndex if it is valid"
        i = startIndex
        while i < len(self.fileList) and i >= 0 :
            if self._isValidIndex(i) :
                return i
            i += 1
        return -1
    def _getCurrentFiles(self) :
        "get a list of files to be displayed"
        self.indicies = []
        self.currentFiles = []
        self.index = self._getValidIndex(self.index)
        i = self.index
        while len(self.indicies) < self.nPlots :
            self.indicies.append(i)
            if i >= 0 :
                self.currentFiles.append(self.fileList[i])
            else :
                self.currentFiles.append(None)
            i = self._getValidIndex(i+1)
        return self.currentFiles
    def run(self) :
        print "Click on plots to (de)select them."
        pylab.show()
    def next(self, event) :
        if VERBOSE >= 2:
            print >> sys.stderr, "Next frame"
        i = self.indicies[-1] + 1
        nextI = self._getValidIndex(i)
        if nextI != -1 :
            self.index = nextI
            self.showData()
    def prev(self, event) :
        if VERBOSE >= 2:
            print >> sys.stderr, "Previous frame"
        if self.index == 0 :
            return
        i = self.index
        n = 0
        while n < self.nPlots and i >= 0 :
            i -= 1
            if self._isValidIndex(i) :
                n += 1
        if i == -1 :
            i = 0
        self.index = i
        self.showData()
    def save(self, event) :
        if VERBOSE >= 2:
            print >> sys.stderr, "Saving"
        saveFileSelections(self.selectedFiles, self.saveFileName)
    def setView(self, event) :
        self.view = event
        self.showData()
    def changeSelection(self, subplotIndex, selection) :
        filename = self.currentFiles[subplotIndex]
        if filename != None :
            self.selectedFiles[filename] = selection
        self.showData()

def main():
    listfilename = sys.argv[1]
    fileSelections = readFileSelections(listfilename)
    fig = UnfoldSelector(selectedFiles=fileSelections, saveFileName=listfilename)
    fig.run()
    return 0

if __name__ == "__main__" :
    main()
