#!/usr/bin/python
#
# unfold_fit - GUI and commandline interface for fitting WLCs to force
# spectroscopy unfolding data and extracting unfolding forces.
#
# Copyright (C) 2008-2009, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.
"""Read/write unfolding force curve fits from/to disk in XML"""

import os.path
from xml.dom import minidom

import model

def XMLnode2dict(node, stringParams=[], boolParams=[], floatParams=[], compositeParams=[], arrayParams=[]) :
    """
    Parse an xlm.dom.minidom node into it's components.  The
    arrayParams option lists elements that may appear multiple times
    (creating an array).  For example
    >>> xml = '\\n'.join(['<element>','  <pi>3.14159</pi>','  <fullname>','    <name>John</name>','    <name>Jacob</name>','    <name>Jingleheimer</name>','    <name>Smith</name>','  </fullname>','</element>'])
    >>> print xml
    <element>
      <pi>3.14159</pi>
      <fullname>
        <name>John</name>
        <name>Jacob</name>
        <name>Jingleheimer</name>
        <name>Smith</name>
      </fullname>
    </element>
    >>> node = minidom.parseString(xml).firstChild
    >>> print node.nodeName
    element
    >>> params = XMLnode2dict(node, floatParams=['pi'], compositeParams=['fullname'])
    >>> print params.keys()
    ['fullname', 'pi']
    >>> print params['pi']
    3.14159
    >>> print params['fullname'].nodeName
    fullname
    >>> fullname = params.pop('fullname')
    >>> ps = XMLnode2dict(fullname, stringParams=['name'], arrayParams=['name'])
    >>> print ps
    {'name': [u'John', u'Jacob', u'Jingleheimer', u'Smith']}
    """
    params = {}
    textBasedParams = [x for x in stringParams] # shallow copy
    textBasedParams.extend(boolParams)
    textBasedParams.extend(floatParams)
    for paramNode in node.childNodes :
        if paramNode.nodeType == node.TEXT_NODE :
            continue
        paramName = str(paramNode.tagName)
        if paramName in compositeParams :
            paramVal = paramNode
        elif paramName in textBasedParams :
            assert len(paramNode.childNodes) == 1, paramNode.childNodes
            assert paramNode.firstChild.nodeType == node.TEXT_NODE
            paramString = ''.join(paramNode.firstChild.data.split()) # uniform whitespace
            if paramName in stringParams :
                paramVal = paramString
            elif paramName in boolParams :
                if paramString == '1' :
                    paramVal = True
                else :
                    paramVal = False
            elif paramName in floatParams :
                paramVal = float(paramString)
        else :
            raise Exception, 'Unknown parameter %s in\n%s' % (paramName, node.toxml())
        if paramName in arrayParams :
            if paramName in params :
                params[paramName].append(paramVal)
            else :
                params[paramName] = [paramVal]
        else :
            assert (paramName in params.keys()) == False, '%s defined twice in:\n%s' % (paramName,node.toxml())
            params[paramName] = paramVal
    return params

def XML2SawtoothModel(xmlstring) :
    """
    Convert an XML string into a SawtoothModel instance
    >>> xml = '\\n'.join(['<unfold>','  <datafile>','    /home/wking/rsrch/data/scaled_unfold/20080919/20080919113008_unfold','  </datafile>','  <selected>0</selected>','  <temperature>298.15</temperature>','  <xScale>1e-9</xScale>','  <yScale>1e-12</yScale>','  <wlc>','    <p>0.4e-9</p>','    <L>102e-9</L>','    <min><x>30.4</x><y>0</y></min>','    <max><x>51</x><y>0</y></max>','  </wlc>','  <wlc>','    <p>0.4e-9</p>','    <L>193e-9</L>','    <min><x>54</x><y>0</y></min>','    <max><x>96</x><y>0</y></max>','  </wlc>','  <tailCut><x>102</x><y>0</y></tailCut>','</unfold>'])
    >>> print xml
    <unfold>
      <datafile>
        /home/wking/rsrch/data/scaled_unfold/20080919/20080919113008_unfold
      </datafile>
      <selected>0</selected>
      <temperature>298.15</temperature>
      <xScale>1e-9</xScale>
      <yScale>1e-12</yScale>
      <wlc>
        <p>0.4e-9</p>
        <L>102e-9</L>
        <min><x>30.4</x><y>0</y></min>
        <max><x>51</x><y>0</y></max>
      </wlc>
      <wlc>
        <p>0.4e-9</p>
        <L>193e-9</L>
        <min><x>54</x><y>0</y></min>
        <max><x>96</x><y>0</y></max>
      </wlc>
      <tailCut><x>102</x><y>0</y></tailCut>
    </unfold>
    >>> model = XML2SawtoothModel(xml)
    >>> print model.datafile
    /home/wking/rsrch/data/scaled_unfold/20080919/20080919113008_unfold
    >>> print model.T
    298.15
    >>> print model.cuts
    [(30.399999999999999, 0.0), (51.0, 0.0), (54.0, 0.0), (96.0, 0.0), (102.0, 0.0)]
    >>> print model.WLClist
    [<WLC p:4e-10 L:1.02e-07 T:298.15>, <WLC p:4e-10 L:1.93e-07 T:298.15>]
    """
    unfoldDoc = minidom.parseString(xmlstring)
    assert len(unfoldDoc.childNodes) == 1, unfoldDoc.childNodes
    unfoldNode = unfoldDoc.firstChild
    assert unfoldNode.tagName == 'unfold', unfoldNode.tagName
    stringParams = ['datafile']
    boolParams = ['selected']
    floatParams = ['temperature', 'xScale', 'yScale', 'xShift', 'yShift']
    compositeParams = ['wlc', 'tailCut']
    arrayParams = ['wlc']
    params = XMLnode2dict(unfoldNode, stringParams, boolParams, floatParams, compositeParams, arrayParams)
    if 'wlc' in params.keys() :
        wlcNodes = params.pop('wlc')
        wlcArray = []
        cuts = []
        stringParams = []
        boolParams = []
        floatParams = ['p', 'L']
        compositeParams = ['min', 'max']
        arrayParams = []
        for node in wlcNodes :
            wlcParams = XMLnode2dict(node, stringParams, boolParams, floatParams, compositeParams, arrayParams)
            min = wlcParams.pop('min')
            minParams = XMLnode2dict(min, floatParams=['x','y'])
            min = (minParams['x'], minParams['y'])
            max = wlcParams.pop('max')
            maxParams = XMLnode2dict(max, floatParams=['x','y'])
            max = (maxParams['x'], maxParams['y'])
            wlcArray.append(model.WLC(**wlcParams))
            cuts.extend([min, max])
        params['WLClist'] = wlcArray
        params['cuts'] = cuts
    if 'tailCut' in params.keys():
        tailCutParams = XMLnode2dict(params.pop('tailCut'),
                                     floatParams=['x','y'])
        tailCut = (tailCutParams['x'], tailCutParams['y'])
        if 'cuts' in params.keys():
            params['cuts'].append(tailCut)
        else:
            params['cuts'] = [tailCut,]
    return model.SawtoothModel(**params)

def bool2XML(tagName, bool, indent="") :
    """
    Convert a boolean value to an XML tag (True->1, False->0)
    >>> print bool2XML("tagname", True, indent="  ")
      <tagname>1</tagname>
    """
    if bool == True :
        val = 1
    else :
        val = 0
    return "%s<%s>%d</%s>" % (indent, tagName, val, tagName)

def SawtoothModel2XML(model) :
    """
    Convert a SawtoothModel instance into an XML string
    >>> m = model.SawtoothModel(datafile="/home/wking/rsrch/data/scaled_unfold/20080919/20080919113008_unfold",selected=False,temperature=298.15,cuts=[(30.4,0),(51,0),(54,0),(96,0),(102,0)],WLClist=[model.WLC(p=0.4e-9,L=103e-9),model.WLC(p=0.4e-9,L=193e-9)])
    >>> m.xscale = 1e-9
    >>> m.yscale = -1e-12
    >>> print SawtoothModel2XML(m)
    <unfold>
      <datafile>
        /home/wking/rsrch/data/scaled_unfold/20080919/20080919113008_unfold
      </datafile>
      <selected>0</selected>
      <temperature>298.15</temperature>
      <xScale>1e-09</xScale>
      <yScale>-1e-12</yScale>
      <xShift>0</xShift>
      <yShift>0</yShift>
      <wlc>
        <p>4e-10</p>
        <L>1.03e-07</L>
        <min>
          <x>30.4</x>
          <y>0</y>
        </min>
        <max>
          <x>51</x>
          <y>0</y>
        </max>
      </wlc>
      <wlc>
        <p>4e-10</p>
        <L>1.93e-07</L>
        <min>
          <x>54</x>
          <y>0</y>
        </min>
        <max>
          <x>96</x>
          <y>0</y>
        </max>
      </wlc>
      <tailCut>
        <x>102</x>
        <y>0</y>
      </tailCut>
    </unfold>
    """
    s = ["<unfold>"]
    s.append("  <datafile>\n    %s\n  </datafile>" % model.datafile)
    s.append(bool2XML("selected", model.selected, indent="  "))
    s.append("  <temperature>%g</temperature>" % model.T)
    s.append("  <xScale>%g</xScale>" % model.xscale)
    s.append("  <yScale>%g</yScale>" % model.yscale)
    s.append("  <xShift>%g</xShift>" % model.xshift)
    s.append("  <yShift>%g</yShift>" % model.yshift)
    if model.WLClist != None :
        for i in range(len(model.cuts)/2):
            wlc = model.WLClist[i]
            min,max = model.getMinMax_raw(i)
            s.append("  <wlc>")
            s.append("    <p>%g</p>" % wlc.p)
            s.append("    <L>%g</L>" % wlc.L)
            s.append("    <min>")
            s.append("      <x>%g</x>" % min[0])
            s.append("      <y>%g</y>" % min[1])
            s.append("    </min>")
            s.append("    <max>")
            s.append("      <x>%g</x>" % max[0])
            s.append("      <y>%g</y>" % max[1])
            s.append("    </max>")
            s.append("  </wlc>")
    if len(model.cuts) % 2 == 1:
        s.append("  <tailCut>")
        s.append("    <x>%g</x>" % model.cuts[-1][0])
        s.append("    <y>%g</y>" % model.cuts[-1][1])
        s.append("  </tailCut>")
    s.append("</unfold>")
    return '\n'.join(s)

def readUnfoldList(filename) :
    """Read an XML unfoldlist into an array of SawtoothModel instances"""
    # code following Dive Into Python
    # http://diveintopython.org/xml_processing/parsing_xml.html
    xmldoc = minidom.parse(filename)
    # XML must have exactly one 'root element' (our 'unfoldlist')
    assert xmldoc.firstChild.tagName == 'unfoldlist'
    unfoldList = []
    for unfoldNode in xmldoc.firstChild.childNodes :
        if unfoldNode.nodeType == unfoldNode.TEXT_NODE :
            continue # ignore text nodes (e.g. whitespace text nodes between the unfold nodes)
        model = XML2SawtoothModel(unfoldNode.toxml())
        unfoldList.append(model)
    return unfoldList

def saveUnfoldList(UnfoldList, filename=None) :
    """Save an array of SawtoothModel instances to an XML file."""
    UnfoldList.sort()

    if filename != None :
        fid = file(filename, 'w')
    else :
        fid = sys.stdout

    print >> fid, '<?xml version="1.0" encoding="UTF-8"?>'
    print >> fid, '<unfoldlist>'
    for model in UnfoldList :
        xml = SawtoothModel2XML(model)
        indentedxml = []
        for line in xml.split('\n') :
            indentedxml.append("  %s" % line)
        indentedxml = '\n'.join(indentedxml)
        print >> fid, indentedxml
    print >> fid, '</unfoldlist>'

    if filename != None :
        fid.close()

def _test() :
    import doctest
    doctest.testmod()

if __name__ == "__main__" :
    _test()
